
#ifndef _CPLRAW_TIMES_HXX_
#define _CPLRAW_TIMES_HXX_ 1

#ifndef _MM_NO_IOSTREAM
#include <iostream>
#endif 

#include <math.h>
#include "cplRaw/failable.hxx"

class xostream;
class xistream;

#if (!( __MINGW32__|| _MSC_VER || MMSYS_lkern))

#include <sys/time.h>

class Time;
class TimerPer;
/*
 * --- class TimeDiff
 */
class TimeDiff {
    double _r;

    double r() const { return _r; } 	/* internal */
    TimeDiff(double r):_r(r) {}
  public:
    TimeDiff(int iSec):_r(iSec) {}

    int iMs() { return int(rint(( _r )*double(1000))); }
    int iUs() { return int(rint(( _r )*double(1000000))); }
  #ifndef _MM_NO_IOSTREAM
    friend std::ostream& operator<<(std::ostream &os, const TimeDiff &) ;
  #endif 

    friend Time operator+( const Time &tiL, const TimeDiff &tdR );
    friend TimeDiff operator-( const Time &tiL, const Time &tiR );
    friend class TimerPer;
};

/*
 *	--- class CalSpec
 *
 *	represents Time values in relation
 *	to a calendar system 
 *
 *	   1999-25-06T15:00
 *	   v
 *	---|--->
 */
class CalSpec:public Failable {
protected:
	struct tm _tm;
	const char *_szTz;			/* timezone variable */
	int _tv_nsec;

public:
	/* Date in ISO 8601 format */
	friend std::ostream& operator<<(std::ostream &os, const CalSpec &) ;
	friend class Time;
//	bool isDef() const { return _fDef; }
//	Time ti();			/* must be defined	*/

//	CalSpec(const struct tm &);	
	CalSpec(const Time &);	
	CalSpec(const Time &, const char *szTz);	
	~CalSpec();
      protected:
	CalSpec(const char *szTz);		 
};

/*
 *	CalSpec1123
 *
 *	outputs calspec in RFC 1123 format
 */
class CalSpec1123:public CalSpec {
public:
	/* Date in RFC 1123 format */
	friend std::ostream& operator<<(std::ostream &os, const CalSpec1123 &) ;

	/*
	 * szTz is not used at the moment. In future releases it will
	 * indicate the timezone the CalSpec object lives in
	 */
	CalSpec1123(const Time &, const char *szTz);	
	CalSpec1123(const Time &);	
	~CalSpec1123();
};

class CalSpecIso:public CalSpec {
public:
	/* Date and time in ISO format YYYYMMDDHHMMSS */
	friend std::ostream& operator<<(std::ostream &os, const CalSpecIso &) ;

	CalSpecIso(const Time &, const char *szTz=0);	
	CalSpecIso(const char *, const char *szTz=0); 
	~CalSpecIso();
};

/*
 *	--- class CalSpecPer
 *
 *	specification of periodic timing in relation to a calendar
 *	system, e.g.:
 *
 *	Daily at 13:21:12
 *	Hourly at -21:12
 *	
 */
class CalSpecPer {
	/*
	 *	spec is 
	 *	((time()-_iOff) mod _iPer) = 0
	 */
	int _iPer;	
	int _iOff;

	friend class TimerPer;
public:
	friend std::ostream& operator<<(std::ostream &os, const CalSpecPer &) ;

	CalSpecPer(int iPer, int iOff=0);
	CalSpecPer(const TimerPer &);	
	~CalSpecPer();
};

/*
 *	class CalRanges
 *
 *	not neccessarily periodic calendar ranges of time
 *
 *	---xxx--x---xxxxxx-------xx-->
 *
 */
class CalRanges {
	/* all inclusive dummy */
	int _fIncl;
public:
	friend std::ostream& operator<<(std::ostream &os, const CalRanges &) ;

	int inRange(const Time &) const { return _fIncl; }
	CalRanges(int fIncl):_fIncl(fIncl) {}		
	~CalRanges() {}
};


/*
 *	Time class, denotes exactly one point in time in
 *	machine specific representation.
 *	
 *	   T
 *	   v
 *	---|--->
 */
class Time {		/* UTC time since epoche begin	*/
	timespec _ts;

	friend class CalSpec;
	friend class TimerPer;
	
	/* internal use only !!! */
	double r() const ; 
	Time(double r);
protected:
	Time();		/* for TimeNil			*/
public:
	friend Time nows();	/* now, second granularity */
	friend Time now();	/* now, best granularity   */
	friend TimeDiff operator-( const Time &tiL, const Time &tiR );
	friend Time operator+( const Time &tiL, const TimeDiff &tdR );
	friend bool operator>=( const Time &tiL, const Time &tiR ); /* l later/equal r */
	
	const timespec &ts() const { return _ts; }
	time_t ti() const { return _ts.tv_sec; }

	/* persistence i/o, 8 bytes */
	friend xostream& operator<<(xostream &os, const Time&) ;
	friend xistream& operator>>(xistream &os, Time&) ;

	/* 
	 * there is no public default constructor for a Time 
	 */
	Time(const Time&); 
	Time(const CalSpec&); 
	Time(const time_t tt);
	Time(const timespec &ts):_ts(ts) {}
	~Time();
};

/*
 *	TimeNil - Time plus NIL value
 */
class TimeNil:private Time {
	int _fDef;		/* defined ?	*/
public:
	bool isDef() const { return (_fDef != 0); }
	const Time &ti() const;	/* if defined	*/

	/* define */
	TimeNil &operator=(const Time &);
	TimeNil &operator=(const TimeNil &);

	/* '-' or CalSpec */
	friend std::ostream& operator<<(std::ostream &os, const TimeNil &) ;

	/* persistence i/o, 12 bytes */
	friend xostream& operator<<(xostream &os, const TimeNil &) ;
	friend xistream& operator>>(xistream &os, TimeNil &) ;

	TimeNil();
	~TimeNil();
};

/*
 *	--- class TimerPer
 *
 *	representation of a periodic Timer, i.e. an open sequence
 *	of equidistant points in time, determined by period duration and 
 *	related to a fixed point is time.
 *
 *	   Ts
 *	   v 
 *	---|-----|-----| - - ->
 *	   ^  d  ^  d  ^
 */
class TimerPer {
	double _rPer;		/* 0 < _rPer		 	*/
	double _rOff;		/* 0 <= _rOffset < _rPer	*/

	friend class CalSpecPer;

protected:
	TimerPer();		/* for TimeNil			*/
public:
	/* persistence i/o, 16 bytes */
	friend xostream& operator<<(xostream &os, const TimerPer&) ;
	friend xistream& operator>>(xistream &os, TimerPer&) ;
	friend int operator==(const TimerPer&tpA, const TimerPer&tpB);

	/*
	 * the points() function is the count of points
	 * between two arbitrary times [tFirst,tLim[.
	 *
	 *
	 *	         V     V
	 *	---|-----|-----|-----+ - - ->
	 *	       ^tF        ^tL
	 */
	int points(const Time&tFirst, const Time &tLim) const;

	/*
 	 * the next function() is the next point in time
	 * >= ti.
	 *
	 *	         V   
	 *	---|-----|- - ->
	 *	       ^ti        
	 */ 
	Time next( const Time &tiRef ) const ;

	/*
	 *	phase synchronous to tiPhase (usually now())
	 *	distance tdDist
	 */
	TimerPer(const Time& tiPhase, const TimeDiff& tdPer);
	TimerPer(const CalSpecPer &);
	~TimerPer();
};	

/*
 *	periodic timer w/ NIL value
 */
class TimerPerNil:private TimerPer {
	int _fDef;		/* defined ?	*/
public:
	bool isDef() const { return (_fDef != 0); }
	const TimerPer &tp() const;	/* if defined	*/

	/* define */
	TimerPerNil &operator=(const TimerPer &);
	TimerPerNil &operator=(const TimerPerNil &);

	/* '-' or CalSpecPer */
	friend std::ostream& operator<<(std::ostream &os, const TimerPerNil &) ;

	/* persistence output, 20 bytes */
	friend xostream& operator<<(xostream &os, const TimerPerNil &) ;
	friend xistream& operator>>(xistream &os, TimerPerNil &) ;
	friend int operator==(const TimerPerNil&tpnA, const TimerPerNil&tpnB);


	TimerPerNil(const TimerPer &);
	TimerPerNil();
	~TimerPerNil();
};





/*
 *	utilities
 */
int isValid(const timespec &ts);	/* ts is well-formed	*/

#endif
#endif
