
#ifndef _GPIB_NECIF_HXX_
#define _GPIB_NECIF_HXX_

#include "gumi_msgs.hxx"
#include "umic/umi_driver.hxx"
#include "gpibc/regs.hxx"
#include "gca.hxx"
#include "cplRaw/sm.hxx"
#include "igcregs.h"
#include "arch.h"
#ifndef _MM_NO_IOSTREAM
 #include <iostream>
#endif

/**
 * \file
 * GPIB conductive driver 
 */

namespace INdepGpib100 {

  namespace Umi=INdepUmi100;

  /**
   * NEC 7210 based interface - Controller and Device 
   */
  class GifControl:public Umi::UmiDriver, private SmM {

   private: 
    /*
     * states
     */
    enum {
      S_POFF,
      S_IDLE,	/* awaiting new command */
      S_CWR,	/* conducted write	*/
      S_CRD,	/* conducted read	*/
      S_CMD,	/* commands		*/
      S_IFC, 	/* send IFC		*/
      S_REN, 	/* send REN false	*/
      S_WAIT, 	/* wait for event	*/
      CNT_STATES
    };



    /* message handlers */
    int msgInit(int iCtx, const Umi::UmiMsgDown *pmd );
    int msgFini(int iCtx, const Umi::UmiMsgDown *pmd);


    int msgIfc(int iCtx, Umi::Pmd pmd);
    int msgRen(int iCtx, Umi::Pmd pmd);
    int msgCmd(int iCtx, Umi::Pmd pmd);
    int msgWr(int iCtx, Umi::Pmd pmd);
    int msgRd(int iCtx, Umi::Pmd pmd);
    int msgSta(int iCtx, Umi::Pmd pmd);
    int msgWait(int iCtx, Umi::Pmd pmd);
    int msgLnReq(int iCtx, Umi::Pmd pmd);
    int msgCsm(int iCtx, Umi::Pmd pmd);



    /* state handlers */
    int completeSend(int iCtx, const GpibTrf *pgm);
    int hCRD(int iCtx);
    int hCWR(int iCtx);
    int hCMD(int iCtx);
    int hIFC(int iCtx);
    int hREN(int iCtx);
    int hWAIT(int iCtx);
    int switchState(int iCtx);

    int gts();
    void cac(bool);
    void setMode( const GpibTermIn &gti);
    void setMode( const GpibTermOut &gto);

    int cSpinCmd() { return 1; }	/* default spin count for commands */
    int cSpinDwr() { return 1; }	/* default spin count for writes   */
    int cSpinDrd() { return 1; }	/* default spin count for reads    */

    /*
     * interface configuration
     */
    GpibIfConf _gic;

    /* protocol instance */
    const Gpib_gco_Proto _pt;


    /*
     * request  parameters
     */
    Umi::Pmd _pmd;		/* current downstream handled	*/
    Umi::Time _tiNext;		/* (T), next timeout		*/
    int _eErr;
    int _iErr;
    void setErr(int eErr, int iErr) { _eErr=eErr, _iErr=iErr; }
    void clrErr() { _eErr=0, _iErr=0; }
    int isErr() { return _eErr; }

    /*
     * IFC, REN variables
     */
    int _fTimerDone;
    
    /*
     * Transfer parameters
     */
    size_t _cTrans;
    size_t _cTransPrev;

    /*
     * READ parameters
     */
    Umi::Pmu _pmuNext;		/* next read response message 	*/
    size_t _cRdRem;		/* remaining read count	      	*/
    GpibTermIn _gti;		/* input terminator		*/
    Umi::Time _tiTmo;		/* (t), per byte timeout	*/
    

    /*
     * enter fail state
     */
    int failed(int iCtx, Umi::Pmd pmd, int eErr, int iErr, int (GifControl::*pf) (int), int eState) ;
    int drop(int iCtx, Umi::Pmd pmd);

    /*
     * responses
     */
    int putUpResp(int iCtx, Umi::Pmu pmu, GpibResp *pmg, int fStat);		/* complete state 	*/	

    /*
     * creates an upstream GpibResp message, fills in the
     * required fields, sends the message upstream
     * and releases pmd.
     */
    int putUpResp( int iCtx, const Umi::UmiMsgDown *pmd , ErrE eErr, int iErr=0  );
    int putUpInfoResp( int iCtx, const Umi::UmiMsgDown *pmd , ErrE eErr, int iErr=0  );	

    /*
     * transactions 
     */
    int _iTransCur;		/* current transaction 		*/
    int _iTransFailed;		/* == iTransCur after fail	*/


    const GpibIfConf &gic() { return _gic; }
    GifCard *pca() const { return static_cast<GifCard*>(Umi::UmiDriver::pca()); }

    /*
     * response utilities
     * release umd, allocate GpibMsgXXX, setup result values and putU
     */
    bool isTransFailed(int i) { return i == _iTransFailed; }
    void failTransaction() { _iTransFailed = _iTransCur; }
    int isTransCur(int i) { return i == _iTransCur; }
    void setTransCur(int i) { _iTransFailed = ~(_iTransCur = i); }


    /* statistics */
    enum StacE {
      STAC_CMD_OUT,
      STAC_BYTE_OUT,
      STAC_BYTE_IN,
      CNT_STAC
    };
    int _mpcStac[CNT_STAC];
    void incStac(StacE e, int c=1) { _mpcStac[e]+=c; }
    void setStac(StacE e, int i) { _mpcStac[e]=i; }


    void debug( const char *sz );

   public:

    /* debug print */
 #ifndef _MM_NO_IOSTREAM
    friend std::ostream &operator<< ( std::ostream &os, const GifControl &);
 #endif

    const Umi::UmiProto *pup(int iChan);		/* proto desc for channel */

    int cStacMax() const { return CNT_STAC; }
    int cStac(int e) const { return _mpcStac[e]; }

    char *print( char *, int ) const;

    /* 
     * message put -- Note: timer() must has been called 
     * once before using putDown() the first time
     */ 
    int putDown(int iCtx, Umi::Pmd pmd, int iChan);

    int openDriver();
    void closeDriver();

    /* interupt handler */
    int interrupt(int iCtx);
    void backenable(int iCtx, int iChan);

    /* timer handler */
    int timer(int iCtx, const Umi::Time &ti);

    GifControl(GifCard *pca);
    virtual ~GifControl();
  };
}
#endif

