#ifndef _UMI_UMI_DRIVER_HXX
#define _UMI_UMI_DRIVER_HXX

/**
 * \file umi_driver.hxx
 * Universal Message Interface - Driver C++ part
 */

#include "umi.hxx"
#include "cplRaw/list.hxx"


namespace INdepUmi100 {

  /**
   * downstream message kernel rep
   */
  class UmiMsgDown:public Link {
    UmiControl *_puc;
    UmiGetBuf *_pug;
   public:
    /** message control part	*/
    const UmiControl *puc() const { return _puc; }
    /** message data part	*/
    UmiGetBuf *pug() const { return _pug; }

    void init(UmiControl *puc, UmiGetBuf *pug) { _puc=puc; _pug=pug; }
    UmiMsgDown():_puc(0), _pug(0) {}
    ~UmiMsgDown() {}
  };

  /**
   * upstream message kernel rep
   */
  class UmiMsgUp:public Link {
    UmiControl *_puc;
    UmiPutBuf *_pup;
   public:
    /** message control part	*/
    UmiControl *puc() const { return _puc; }
    /** message data part	*/
    UmiPutBuf *pup() const { return _pup; }

    void init(UmiControl *puc, UmiPutBuf *pup) { _puc=puc; _pup=pup; }
    UmiMsgUp():_puc(0), _pup(0) {}
    ~UmiMsgUp() {}
  };

  typedef const UmiMsgDown *Pmd;
  typedef const UmiMsgUp *Pmu;


  /**
   * self-contained downstream message copy
   */
  class UmiMsgDownBuf:public UmiMsgDown, public UmiGetBuf, public UmiControlBuf { 
    char _mpchData[ __SZ_DATA_MAX ];
   public:
    #ifndef _MM_NO_IOSTREAM
     friend std::ostream &operator<<(std::ostream &os, UmiMsgDownBuf &ud);
    #endif

    /*
     * copy in user messages
     */
    UmiControl& operator=(const UmiControl &uc) { return UmiControlBuf::operator=(uc); }

    UmiMsgDownBuf() ;
    ~UmiMsgDownBuf() ;
  };

  /**
   * self-contained downstream message copy
   */
  class UmiMsgUpBuf:public UmiMsgUp, public UmiPutBuf, public UmiControlBuf {
    char _mpchData[ __SZ_DATA_MAX ];
   public:
    #ifndef _MM_NO_IOSTREAM
     friend std::ostream &operator<<(std::ostream &os, UmiMsgUpBuf &ud);
    #endif

    UmiMsgUpBuf();
    ~UmiMsgUpBuf();
  };

  /**
   * user/driver copy
   */
  void copy(UmiData *, const UmiMsgUpBuf*);
  void copy(UmiMsgDownBuf*, const UmiData*);

  class UmiChannel;
  /**
   * upper message handler interface definition
   */
  class UmiEnv {
   public:

    /*
     *  tiNext() schedules a call to UmiDriver::timer()
     *  not earlier than the "current time" + tiDelta (the target time).
     *
     *  It returns the target time, which must be
     *  used by timer() to qualify the time at call as
     *  being equal or later than than the target time.
     *
     *  tiDelta is limited to tiLim(). 
     *  
     */
    virtual Time tiNext( const Time &tiDelta )=0;		

    /**
     * "get ready" signal function
     *
     * sigGet() is called by the driver when there is a
     * new message ready for pickup at the upstream queue.
     * The derived class may use the signal to continue
     * any thread waiting for upstream data.
     * The default implementation of sigGet() does nothing.
     *
     * \arg puc is the channel getting ready.
     * You can set and use an enviroment specific handle
     * for each channel via UmiChannel::setHandle(void *p). Alternatively,
     * UmiChannel::iChan() may be used to fetch the channel's number (starting at 0 for the first channel).
     *
     * \arg iCtx is the context identifier (i.e. the iCtx argument of the top level function resulting
     * in a call to setGet()).
     */
    virtual void sigGet( int iCtx, UmiChannel *puc );		

    /**
     * "put ready" signal function
     *
     * sigPut() is called by the driver when there is room
     * at the downstream queue to accommodate at least one
     * downstream message. 
     * The default implementation of sigPut() does nothing.
     * For argument descriptions see sigGet().
     */
    virtual void sigPut( int iCtx, UmiChannel *puc );		

    UmiEnv() {}
    virtual ~UmiEnv() {}
  };

  class UmiDriver;

  /**
   * driver channel
   */
  class UmiChannel : public Link {
    /* driver */
    UmiDriver*  _pud;
    UmiEnv*	_pue;

    /* channel number, identifies channel @ driver */
    int	        _iChan;		/* chan @ _pud			*/
    void*	_phEnv;		/* handle at _pue		*/

    /*
     * message pools
     */
    ChainRaw _cnUmuPool;
    ChainRaw _cnUmdPool;

    /*
     * downstream messages
     */
    ChainRaw _cnUmdQueue; 			/* messages queued for free channels 	*/
    ChainRaw _cnUmdDriver;			/* per channel message (single)		*/
    ChainRaw _cnUmdUser;			/* in process by user			*/

    /* 
     * upstream messages 
     */
    ChainRaw _cnUmuQueue;			/* sent upstream 			*/
    ChainRaw _cnUmuDriver;			/* in process by driver			*/
    ChainRaw _cnUmuUser;			/* in process by user			*/

    ChainRaw *pcnUmuPool() { return &_cnUmuPool; }
    ChainRaw *pcnUmdPool() { return &_cnUmdPool; }
    ChainRaw *pcnUmdUser() { return &_cnUmdUser; }	

    ChainRaw *pcnUmuDriver() { return &_cnUmuDriver; }	/* allocated messages	*/
    ChainRaw *pcnUmuQueue() { return &_cnUmuQueue; }	/* msgs sent upstream   */
    ChainRaw *pcnUmuUser() { return &_cnUmuUser; }	

    ChainRaw *pcnUmdQueue() { return &_cnUmdQueue; }
    ChainRaw *pcnUmdDriver() { return &_cnUmdDriver; }
   public:
    /* init buffers  */
    void initPool(UmiMsgDownBuf *pUmd) { _cnUmdPool.push(pUmd); }
    void initPool(UmiMsgUpBuf *pUmu) { _cnUmuPool.push(pUmu); }

    /* move all back to the pools and call finiPool()	*/
    void finiQueues();			
    void finiPools();			
    virtual void finiPool(Link *);

    #ifndef _MM_NO_IOSTREAM
     std::ostream &printQueues(std::ostream &os) const ;
    #endif 

    UmiDriver *pud() { return _pud; }		/**< this channels driver	*/
    UmiEnv *pue() { return _pue; }		/**< get the channels environment */
    int iChan() { return _iChan; }		/**< get the channels number	*/

    /**
     * open of a driver channel. 
     *
     * Does call \ref UmiDriver::open() for this channel.
     */
    virtual int open();

    /**
     * close of a driver channel. 
     *
     * Does call \ref UmiDriver::close() for this channel and then
     * dismantels all buffers associated with that channel.
     */
    virtual void close();

    /**
     * Upstream message available if nonzero.
     */
    int pollIn()  { return pcnUmuQueue()->c(); } 
    UmiMsgUpBuf*getUpBuf();		/**< get from up queue, iff pollIn()	*/
    void releaseUpBuf(int iCtx, UmiMsgUpBuf *);	/**< release most recent got from getUpBuf()	*/

    /**
     * Space for at least one downstream message if nonzero.
     */
    int pollOut() { return pcnUmdPool()->c(); } 
    UmiMsgDownBuf* allocDownBuf();	/**s allocate a downstream buffer	*/
    void putDownBuf(int iCtx, UmiMsgDownBuf*);	/**< put downstream buffer to driver		*/

    /**
     * send a message upstream, message and buffer must have been obtained
     * by allocUp() and are released be this call
     */
    virtual void putUp(int iCtx, Pmu );

    /**
     * allocate an upstream message
     */
    virtual Pmu allocUp(size_t sData);

    /**
     * allocate an upstream message. If the queue is
     * blocked, remove the last message thas has been put upstream via putUp() and reallocate it.
     */
    virtual Pmu allocUpForced(size_t sData);

    /**
     * allocate an upstream message. Declares that the control
     * and data buffers of pmd are no more used and may be
     * reused for the upstream message. 
     */
    virtual Pmu allocUp(size_t sData, Pmd pmd);


    /**
     * release an downstream message, ANY downstream message must
     * be released this way. Clears channel. 
     * \note UmiDriver::putDown() MAY BE CALLED DURING releaseDown() FOR THE CHANNEL.
     */
    virtual void releaseDown(int iCtx, Pmd);

    /** 
     * next timer tick >= now + ti (depends on system clock granularity)
     */
    Time tiNext( const Time &ti ) { return pue()->tiNext(ti); }


    const UmiProto *pup(); 

    /**
     * the the environment specific handle for this channel.
     * The default handle is 0.
     */
    void* getHandle() { return _phEnv; }

    /**
     * set the enviroment specific handle for this channel. 
     */
    void setHandle(void *p) { _phEnv=p; }
    void setDriver(UmiDriver *pud) { _pud=pud; }
    void setChan(int i) { _iChan=i; }
    UmiChannel(UmiDriver *pud, int iChan, UmiEnv *pue);
    virtual ~UmiChannel();
  };

  /**
   * timer interface def
   */
  class UmiCard;

  /**
   * lower message handler (driver) base class 
   */
  class UmiDriver : public Link {
   public:
    typedef int DownE;
   private:
    const char *_szName;
    UmiChannel *_pcn;
    UmiCard *_pca;
    ChainRaw _cnuc;			/* channel list		*/
    int _fOpen;				/* open channel flag	*/
    friend int UmiChannel::open();
    friend void UmiChannel::close();

    /**
     * open of a driver - base class
     *
     * Is called by the channel open for each channel.
     *
     * If the current call is the first open of a channel of this driver, 
     * the virtual openDriver() is called after successful opening the base class.
     *
     * In any case, the openChannel() is called for the derived class.
     */
    int open(int iChan);

    /**
     * close of a driver - base class
     *
     * Is called by the channel close for each channel.
     *
     * In any case, closeChannel() is called for the derived class.
     *
     * If the current call is the last close of a channel of this driver,
     * The virtual closeDriver() is called before the finilization of the base class.
     */
    void close(int iChan);

   protected:
    Time tiNext( const Time &ti ) { return pcn(0)->tiNext(ti); }

   public:
    UmiChannel *pcn(int i) { return static_cast<UmiChannel*>(_cnuc.get(i)); }
    int ccn() { return _cnuc.c(); }

    void startIterChannels( ChainIterator *pit) { _cnuc.strtIter(pit); }
    UmiChannel *iterChannels( ChainIterator *pit) { return static_cast<UmiChannel*>(_cnuc.iter(pit)); }

    UmiCard *pca() const { return _pca; }
    const char *szName() { return _szName ; }		/* driver name, qualified by card's name */


    /** 
     * downstream message put for channel \ref iChan.
     * The buffer may be reused for
     * upstream message transfer, see UmiChannel::allocUp().
     * The message channel is locked until UmiChannel::releaseDown() is called.
     */ 
    virtual DownE  putDown(int iCtx, Pmd pmd, int iChan)=0;

    /**
     * per channel (driver?) protocol descriptor
     */
    virtual const UmiProto *pup(int iChan)=0;

    /** 
     * open of a driver's channel - derived class 
     *
     * The derived class may implement this member function
     * if it requires notification on the opening of a channel.
     */
    virtual int openChannel( int iChan );

    /** 
     * close of a driver's channel - derived class
     *
     * The derived class may implement this member function
     * if it requires notification on the closing of a channel.
     */
    virtual void closeChannel( int iChan );

    int isOpen();	/**< is this driver open ? */

    /**
     * open of a driver - derived class
     *
     * Any derived function may enable interrupt sources. 
     */
    virtual int openDriver();

    /**
     * close of a driver - derived class
     *
     * Any derived function must disable interrupt sources.
     */
    virtual void closeDriver();

    virtual char *print( char *, int ) const;

    /**
     * channel back-enable 
     */
    virtual void backenable(int iCtx, int iChan);

    /** interupt handler */
    virtual int interrupt(int iCtx)=0;

    /** timer handler */
    virtual int timer(int iCtx, const Time &)=0;

    /** max size of message data part */
    size_t sizeDataMax() const { return __SZ_DATA_MAX; }


    void init(UmiChannel *pcn) { _cnuc.push(pcn); }
    void fini(UmiChannel *pcn) { _cnuc.remove(pcn); }
    UmiDriver(const char *szName, UmiCard *pca);
    virtual ~UmiDriver();
  };

  /**
   * Card abstraction
   */
  class UmiCard {
    ChainRaw _cnud;		/* all drivers		*/
    UmiDriver *_pudCur;		/* "current" driver	*/
   protected:
   public:
    UmiDriver *pudCur() { return _pudCur; }
    int setDriver(UmiDriver *pud) { if (_pudCur) return -1; _pudCur=pud; return 0; }
    int clrDriver(UmiDriver *pud) { if (_pudCur != pud) return -1; _pudCur=0; return 0; }

    void startIterDrivers( ChainIterator *pit) { _cnud.strtIter(pit); }
    UmiDriver *iterDrivers( ChainIterator *pit) { return static_cast<UmiDriver*>(_cnud.iter(pit)); }



    /** 
     * per card primary interrupt function.
     * To be called from the enviroment whenever an interrupt from this
     * card might have been occured. As interrupts may be shared, a call
     * of interrupt() might have been done for another card.
     */
    virtual int interrupt(int iCtx)=0;

    /** per card "hard" timer */
    virtual int timer(int iCtx, const Time &)=0;

    virtual char *print( char *, int ) const ;

    void initDriver(UmiDriver *pud) { _cnud.push(pud); }
    UmiCard():_pudCur(0) {}
    virtual ~UmiCard() {}
  };

}  


#endif


